# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

from causal_hub.models import CatBN, CatCTBN, GaussBN

def load_alarm() -> CatBN:
    r"""
    Load the `ALARM` categorical BN from the assets.
    """

def load_andes() -> CatBN:
    r"""
    Load the `ANDES` categorical BN from the assets.
    """

def load_arth150() -> GaussBN:
    r"""
    Load the `ARTH150` Gaussian BN from the assets.
    """

def load_asia() -> CatBN:
    r"""
    Load the `ASIA` categorical BN from the assets.
    """

def load_barley() -> CatBN:
    r"""
    Load the `BARLEY` categorical BN from the assets.
    """

def load_cancer() -> CatBN:
    r"""
    Load the `CANCER` categorical BN from the assets.
    """

def load_child() -> CatBN:
    r"""
    Load the `CHILD` categorical BN from the assets.
    """

def load_diabetes() -> CatBN:
    r"""
    Load the `DIABETES` categorical BN from the assets.
    """

def load_earthquake() -> CatBN:
    r"""
    Load the `EARTHQUAKE` categorical BN from the assets.
    """

def load_eating() -> CatCTBN:
    r"""
    Load the `EATING` categorical CTBN from the assets.
    """

def load_ecoli70() -> GaussBN:
    r"""
    Load the `ECOLI70` Gaussian BN from the assets.
    """

def load_hailfinder() -> CatBN:
    r"""
    Load the `HAILFINDER` categorical BN from the assets.
    """

def load_hepar2() -> CatBN:
    r"""
    Load the `HEPAR2` categorical BN from the assets.
    """

def load_insurance() -> CatBN:
    r"""
    Load the `INSURANCE` categorical BN from the assets.
    """

def load_link() -> CatBN:
    r"""
    Load the `LINK` categorical BN from the assets.
    """

def load_magic_irri() -> GaussBN:
    r"""
    Load the `MAGIC_IRRI` Gaussian BN from the assets.
    """

def load_magic_niab() -> GaussBN:
    r"""
    Load the `MAGIC_NIAB` Gaussian BN from the assets.
    """

def load_mildew() -> CatBN:
    r"""
    Load the `MILDEW` categorical BN from the assets.
    """

def load_munin1() -> CatBN:
    r"""
    Load the `MUNIN1` categorical BN from the assets.
    """

def load_pathfinder() -> CatBN:
    r"""
    Load the `PATHFINDER` categorical BN from the assets.
    """

def load_pigs() -> CatBN:
    r"""
    Load the `PIGS` categorical BN from the assets.
    """

def load_sachs() -> CatBN:
    r"""
    Load the `SACHS` categorical BN from the assets.
    """

def load_survey() -> CatBN:
    r"""
    Load the `SURVEY` categorical BN from the assets.
    """

def load_water() -> CatBN:
    r"""
    Load the `WATER` categorical BN from the assets.
    """

def load_win95pts() -> CatBN:
    r"""
    Load the `WIN95PTS` categorical BN from the assets.
    """
