/// Chunk length used for axis-wise parallel processing.
pub const AXIS_CHUNK_LENGTH: usize = 4096;
/// Epsilon value used for floating-point comparisons.
pub const EPSILON: f64 = 1e-8;
/// ln(2 * PI).
#[allow(clippy::excessive_precision)]
pub const LN_2_PI: f64 = 1.83787706640934533908193770912475883960723876953125;
